import { __decorate } from "tslib";
import { html, css, LitElement, nothing } from 'lit';
import { customElement, property, state } from 'lit/decorators.js';
import { env, getDestinationWidthBySize, } from '@responsive-image/core';
import { classMap } from 'lit/directives/class-map.js';
import { ifDefined } from 'lit/directives/if-defined.js';
import { decode2url } from '@responsive-image/core/blurhash/decode';
import { styleMap } from 'lit/directives/style-map.js';
var Layout;
(function (Layout) {
    Layout["RESPONSIVE"] = "responsive";
    Layout["FIXED"] = "fixed";
})(Layout || (Layout = {}));
const PIXEL_DENSITIES = [1, 2];
// determines the order of sources, prefereing next-gen formats over legacy
const typeScore = new Map([
    ['png', 1],
    ['jpeg', 1],
    ['webp', 2],
    ['avif', 3],
]);
let ResponsiveImage = class ResponsiveImage extends LitElement {
    constructor() {
        super(...arguments);
        this.loading = 'lazy';
        this.decoding = 'async';
        this.alt = '';
        this.complete = false;
    }
    get layout() {
        return this.width === undefined && this.height === undefined
            ? Layout.RESPONSIVE
            : Layout.FIXED;
    }
    get sources() {
        if (this.layout === Layout.RESPONSIVE) {
            return this.src.imageTypes.map((type) => {
                let widths = this.src.availableWidths;
                if (!widths) {
                    widths = env.deviceWidths;
                }
                const sources = widths.map((width) => {
                    const url = this.src.imageUrlFor(width, type);
                    return `${url} ${width}w`;
                });
                return {
                    srcset: sources.join(', '),
                    sizes: this.sizes ?? (this.size ? `${this.size}vw` : undefined),
                    type,
                    mimeType: `image/${type}`,
                };
            });
        }
        const { width } = this;
        if (width === undefined) {
            return [];
        }
        return this.src.imageTypes.map((type) => {
            const sources = PIXEL_DENSITIES.map((density) => {
                const url = this.src.imageUrlFor(width * density, type);
                return `${url} ${density}x`;
            }).filter((source) => source !== undefined);
            return {
                srcset: sources.join(', '),
                type,
                mimeType: `image/${type}`,
            };
        });
    }
    get sourcesSorted() {
        return this.sources.sort((a, b) => (typeScore.get(b.type) ?? 0) - (typeScore.get(a.type) ?? 0));
    }
    get imgWidth() {
        if (this.layout === Layout.RESPONSIVE) {
            return getDestinationWidthBySize(this.size);
        }
        if (this.width) {
            return this.width;
        }
        const ar = this.src.aspectRatio;
        if (ar !== undefined && ar !== 0 && this.height !== undefined) {
            return this.height * ar;
        }
        return undefined;
    }
    get imgHeight() {
        if (this.height) {
            return this.height;
        }
        const ar = this.src.aspectRatio;
        if (ar !== undefined && ar !== 0 && this.imgWidth !== undefined) {
            return Math.round(this.imgWidth / ar);
        }
        return undefined;
    }
    get imgSrc() {
        return this.src.imageUrlFor(this.imgWidth ?? 640);
    }
    get hasLqipBlurhash() {
        return this.src.lqip?.type === 'blurhash';
    }
    get showLqipBlurhash() {
        return !this.complete && this.hasLqipBlurhash;
    }
    get lqipBlurhash() {
        if (!this.hasLqipBlurhash) {
            return undefined;
        }
        const { hash, width, height } = this.src.lqip;
        const uri = decode2url(hash, width, height);
        return `url("${uri}")`;
    }
    render() {
        const { lqip } = this.src;
        const classes = {
            'ri-responsive': this.layout === Layout.RESPONSIVE,
            'ri-fixed': this.layout === Layout.FIXED,
            'ri-lqip-inline': lqip?.type === 'inline' && !this.complete,
            'ri-lqip-color': lqip?.type === 'color' && !this.complete,
            'ri-lqip-blurhash': lqip?.type === 'blurhash' && !this.complete,
            [lqip?.type === 'color' || lqip?.type === 'inline' ? lqip.class : '']: (lqip?.type === 'color' || lqip?.type === 'inline') && !this.complete,
        };
        const styles = this.showLqipBlurhash
            ? {
                backgroundImage: this.lqipBlurhash,
                backgroundSize: 'cover',
            }
            : {};
        return html `
      <picture>
        ${this.sourcesSorted.map((s) => html `<source
              srcset=${s.srcset}
              type=${s.mimeType}
              sizes=${s.sizes ?? nothing}
            />`)}
        <img
          part="img"
          width=${ifDefined(this.imgWidth)}
          height=${ifDefined(this.imgHeight)}
          class=${classMap(classes)}
          style=${styleMap(styles)}
          src=${ifDefined(this.imgSrc)}
          alt=${this.alt}
          loading=${this.loading}
          decoding=${this.decoding}
          crossorigin=${ifDefined(this.crossOrigin)}
          fetchpriority=${ifDefined(this.fetchPriority)}
          referrerpolicy=${ifDefined(this.referrerPolicy)}
          @load=${(event) => {
            this.complete = true;
            this.dispatchEvent(new Event(event.type, event));
        }}
          @error=${(event) => {
            this.dispatchEvent(new ErrorEvent(event.type, event));
        }}
          @abort=${(event) => {
            this.dispatchEvent(new Event(event.type, event));
        }}
        />
      </picture>
    `;
    }
};
ResponsiveImage.styles = css `
    .ri-responsive {
      width: 100%;
      height: auto;
    }

    .ri-fixed,
    .ri-responsive {
      content-visibility: auto;
    }

    .ri-lqip-inline {
      background-size: cover;
    }
  `;
__decorate([
    property({ type: Object })
], ResponsiveImage.prototype, "src", void 0);
__decorate([
    property({ type: Number })
], ResponsiveImage.prototype, "size", void 0);
__decorate([
    property({ type: String })
], ResponsiveImage.prototype, "sizes", void 0);
__decorate([
    property({ type: Number })
], ResponsiveImage.prototype, "width", void 0);
__decorate([
    property({ type: Number })
], ResponsiveImage.prototype, "height", void 0);
__decorate([
    property({ type: String })
], ResponsiveImage.prototype, "loading", void 0);
__decorate([
    property({ type: String })
], ResponsiveImage.prototype, "decoding", void 0);
__decorate([
    property({ type: String })
], ResponsiveImage.prototype, "fetchPriority", void 0);
__decorate([
    property({ type: String })
], ResponsiveImage.prototype, "crossOrigin", void 0);
__decorate([
    property({ type: String })
], ResponsiveImage.prototype, "referrerPolicy", void 0);
__decorate([
    property({ type: String })
], ResponsiveImage.prototype, "alt", void 0);
__decorate([
    state()
], ResponsiveImage.prototype, "complete", void 0);
ResponsiveImage = __decorate([
    customElement('responsive-image')
], ResponsiveImage);
export { ResponsiveImage };
//# sourceMappingURL=responsive-image.js.map